/*  plumb - the pipe shell - split/merge streams
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "process.h"
#include "fd.h"
#include "plumb.h"
#include "virt.h"

static virt_ctrl_t hub_close_on_eof(proc_virt_t *proc, int local_fd)
{
	int n;

	/* pass on eof and exit */
	for(n = 0; n < proc->h.num_fds; n++)
		fd_close(&proc->h.fds[n]);

	return VC_QUIT;
}

virt_ctrl_t hub_source_block(proc_virt_t *proc, int local_fd, int is_blocking)
{
	int n, num_blocks;

	/* check all sources - if any of them block, we are blocked */
	num_blocks = 0;
	for(n = 0; n < proc->h.num_fds; n++)
		if (is_source(&proc->h.fds[n]))
			if (vp_is_blocked(proc, n))
				num_blocks++;

	if (
	       ((is_blocking == 0) && (num_blocks == 0))  /* release blocks: is_blocking == 0 and this was the last blocking source */
	    || ((is_blocking == 1) && (num_blocks == 1))  /* set block: is_blocking == 1 and we have exactly 1 blocking source (means thisis the first) */
			)
	{
		for(n = 0; n < proc->h.num_fds; n++)
			if (is_sink(&proc->h.fds[n]))
				back_block(&proc->h.fds[n], is_blocking);
	}

	return VC_OK;
}

virt_ctrl_t hub_source_eof(proc_virt_t *proc, int local_fd)
{
	/* if a blocking source is closed, it is not blocking any more */
	proc->h.fds[local_fd].blocked = 0;     /* will be also set by our caller but it's essential that it's already 0 when we call the block accounting */
	return hub_source_block(proc, local_fd, 0);
}

static virt_ctrl_t hub_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	int n;

	/* pass on data */
	for(n = 0; n < proc->h.num_fds; n++)
		if (is_source(&proc->h.fds[n]))
			fd_write(&proc->h.fds[n], data, len);

	return VC_OK;
}

int vp_hub_init(proc_virt_t *p, kvlist_t *args)
{
	kv_t *k;

	/* allocate a process descriptor */
	p->input        = hub_input;
	p->sink_eof     = hub_close_on_eof;
	p->source_eof   = hub_close_on_eof;
	p->source_block = hub_source_block;
	p->timer = NULL;
	p->start = NULL;
	p->data = NULL;

	if (args != NULL) {
		for(k = args->head; k != NULL; k = k->next) {
			if (strcmp(k->key, "eof") == 0) {
				if (strcmp(k->value, "ignore") == 0) {
					p->source_eof = hub_source_eof; /* for maintaining blocks */
					p->sink_eof   = NULL;           /* silently ignore eof on sink - fd_write() will fail, and we will ignore that */
				}
				else if (strcmp(k->value, "ignore-on-source") == 0)
					p->source_eof = hub_source_eof; /* for maintaining blocks */
				else if (strcmp(k->value, "ignore-on-sink") == 0)
					p->sink_eof   = NULL;           /* silently ignore eof on sink - fd_write() will fail, and we will ignore that */
			}
			else
				plumb_error("Invalid argument for hub: %s\n", k->key);
		}
	}

	return 0;
}
